use bevy::sprite::MaterialMesh2dBundle;
use bevy::sprite::Mesh2dHandle;
use bevy::{prelude::*, window::PrimaryWindow};

use crate::camera::CameraTarget;
use crate::{
    assets::AssetLoader,
    movement::{Heading, Mover, Speed, Velocity},
    statue::Statue,
};

const PLAYER_SPEED: f32 = 1000.;

#[derive(Component, Debug)]
pub struct Player;

#[derive(Component, Debug)]
pub struct Crosshair;

impl Plugin for Player {
    fn build(&self, app: &mut App) {
        app.add_systems(PostStartup, spawn_player)
            .add_systems(Update, (sprite_select, controls.after(sprite_select)));
    }
}

fn spawn_player(
    mut commands: Commands,
    assets: Res<AssetLoader>,
    asset_server: Res<AssetServer>,
    mut meshes: ResMut<Assets<Mesh>>,
    mut materials: ResMut<Assets<ColorMaterial>>,
) {
    let texture = asset_server.load("img/Player.png");
    let listener = SpatialListener::new(1.);

    commands
        .spawn((
            Player,
            CameraTarget,
            SpriteSheetBundle {
                texture,
                atlas: assets.images.player.clone().into(),
                sprite: Sprite::default(),
                ..default()
            },
            Mover {
                velocity: Velocity(Vec2::ZERO),
                heading: Heading(Vec2::ZERO),
                speed: Speed(PLAYER_SPEED),
            },
        ))
        .with_children(|player| {
            player.spawn((SpatialBundle::default(), listener.clone()));
            player.spawn(TextBundle {
                text: Text::from_section(
                    "You",
                    TextStyle {
                        font_size: 32.0,
                        font: assets.fonts.iosevkalytemin.clone(),
                        color: Color::WHITE,
                        ..Default::default()
                    },
                ),
                transform: Transform::from_translation(Vec3::new(30.0, -30.0, -1000.0)),
                ..Default::default()
            });

            let mesh = Mesh2dHandle(meshes.add(Capsule2d::new(3.0, 25.0)));
            let material = materials.add(Color::hsl(360. * 1 as f32 / 3 as f32, 0.95, 0.7));
            // let global_transform = GlobalTransform::from_xyz(
            //     0.0, 20.0,
            //     // TODO: need some way to ensure this draws above everything else?
            //     // a UI layer or something?
            //     1000.,
            // );

            let transform = Transform::from_xyz(
                0.0, 0.0,
                // TODO: need some way to ensure this draws above everything else?
                // a UI layer or something?
                1000.,
            );

            player.spawn((
                Crosshair,
                MaterialMesh2dBundle {
                    // global_transform,
                    mesh,
                    material,
                    transform,
                    ..default()
                },
            ));
        });
}

fn controls(
    mut commands: Commands,
    mut player_q: Query<(&mut Sprite, &mut TextureAtlas, &mut Heading, Entity), With<Player>>,
    asset_server: Res<AssetServer>,
    input: Res<ButtonInput<KeyCode>>,
    mouse_input: Res<ButtonInput<MouseButton>>,
    assets: Res<AssetLoader>,
    win_q: Query<&Window, With<PrimaryWindow>>,
    cam_q: Query<(&Camera, &GlobalTransform), With<Camera>>,
    mut crosshair_q: Query<&mut Transform, With<Crosshair>>,
) {
    let (mut sprite, mut texture, mut heading, player_entity) = player_q.single_mut();
    let (camera, camera_transform) = cam_q.single();

    let mouse_world_position = win_q
        .single()
        .cursor_position()
        .and_then(|cursor| camera.viewport_to_world(&GlobalTransform::from_xyz(0., 0., 0.), cursor))
        .map(|ray| ray.origin.truncate());

    if let Some(pos) = mouse_world_position {
        if let Ok(mut t) = crosshair_q.get_single_mut() {
            t.look_at(Vec3::ZERO, pos.extend(0.));
            t.translation = camera_transform.translation();
        }
    }

    **heading = Vec2::ZERO;
    for button in mouse_input.get_just_pressed() {
        match button {
            MouseButton::Left => {
                if let Some(world_position) = win_q
                    .single()
                    .cursor_position()
                    .and_then(|cursor| camera.viewport_to_world(camera_transform, cursor))
                    .map(|ray| ray.origin.truncate())
                {
                    let spos = world_position.extend(-world_position.y);
                    let statue_texture = asset_server.load("img/Props.png").clone();
                    commands.spawn((
                        Statue,
                        SpriteSheetBundle {
                            texture: statue_texture,
                            atlas: TextureAtlas {
                                layout: assets.images.statue.clone(),
                                index: 0,
                            },
                            transform: Transform::from_translation(spos),
                            ..Default::default()
                        },
                    ));
                }
            }
            _ => {}
        }
    }
    for key in input.get_just_pressed() {
        match key {
            KeyCode::Space => {
                let child = commands
                    .spawn(AudioSourceBundle {
                        source: assets.sounds.meow.clone(),
                        settings: PlaybackSettings::DESPAWN.with_spatial(false),
                    })
                    .id();
                commands.entity(player_entity).push_children(&[child]);
            }
            _ => {}
        }
    }
    for key in input.get_pressed() {
        match key {
            KeyCode::KeyA | KeyCode::ArrowLeft => {
                **heading -= Vec2::X;
            }
            KeyCode::KeyD | KeyCode::ArrowRight => {
                **heading += Vec2::X;
            }
            KeyCode::KeyW | KeyCode::ArrowUp => {
                **heading += Vec2::Y;
            }
            KeyCode::KeyS | KeyCode::ArrowDown => {
                **heading -= Vec2::Y;
            }
            _ => {}
        }
    }
    if heading.y < 0. {
        texture.index = 0;
    } else if heading.y > 0. {
        texture.index = 1;
    } else if heading.x != 0. {
        texture.index = 2;
        sprite.flip_x = heading.x > 0.;
    }
}

fn sprite_select(
    mut q: Query<(&Handle<TextureAtlasLayout>, &mut TextureAtlas), With<Player>>,
    keyboard_input: Res<ButtonInput<KeyCode>>,
    texture_atlases: Res<Assets<TextureAtlasLayout>>,
) {
    if let Ok((texture, mut sprite)) = q.get_single_mut() {
        if keyboard_input.just_pressed(KeyCode::KeyN) {
            let t = texture_atlases
                .get(texture)
                .expect("could not load player texture");
            if sprite.index < t.textures.len() - 1 {
                sprite.index += 1;
            } else {
                sprite.index = 0;
            }
        }
    }
}
